@@ -25,6 +25,7 @@ def get_cloud_json_from_id(cloud_id: str, api_client: DefaultApi) -> Dict["str",
         "provider": cloud.provider,
         "region": cloud.region,
         "credentials": cloud.credentials,
+        "config": cloud.config,
     }
 
 
@@ -65,7 +66,7 @@ def get_cloud_id_and_name(
         if len(clouds) > 1:
             raise ClickException(
                 "Multiple clouds: {}\n"
-                "Please specify the one you want to refer to.".format(
+                "Please specify the one you want to refer to using --cloud-name.".format(
                     [cloud.name for cloud in clouds]
                 )
             )
