@@ -1,9 +1,12 @@
 import hashlib
+import inspect
+import json
 import os
 import subprocess
 import sys
+import tempfile
 import time
-from typing import Any, Dict, List, Optional, Tuple, Union
+from typing import Any, cast, Dict, List, Optional, Tuple, Union
 
 import colorama  # type: ignore
 import ray
@@ -26,6 +29,17 @@ ANYSCALE_FILES_HASH = "anyscale.com/local_files_hash"
 # Max number of auto created sessions.
 MAX_SESSIONS = 20
 
+# Default minutes for autosuspend.
+DEFAULT_AUTOSUSPEND_TIMEOUT = 120
+
+
+# Default docker images to use for connect sessions.
+def _get_base_image(cpu_or_gpu: str = "cpu") -> str:
+    py_version = "".join(str(x) for x in sys.version_info[0:2])
+    if py_version not in ["36", "37", "38"]:
+        raise ValueError("No default docker image for py{}".format(py_version))
+    return "anyscale/ray-ml:nightly-py{}-{}".format(py_version, cpu_or_gpu)
+
 
 class _ConsoleLog:
     def __init__(self) -> None:
@@ -59,6 +73,18 @@ class _ConsoleLog:
             )
             print(*msg)
 
+    def error(self, *msg: str) -> None:
+        print(
+            "{}{}(anyscale +{}){} ".format(
+                colorama.Style.BRIGHT,
+                colorama.Fore.RED,
+                self._time_string(),
+                colorama.Style.RESET_ALL,
+            ),
+            end="",
+        )
+        print(*msg)
+
     def _time_string(self) -> str:
         delta = time.time() - self.t0
         hours = 0
@@ -145,11 +171,14 @@ class SessionBuilder:
 
         # Builder args.
         self._scratch_dir: str = scratch_dir
-        self._ray_release: str = "latest"
+        self._ray_release: Optional[str] = None
         self._project_dir: Optional[str] = None
+        self._cloud_name: Optional[str] = None
         self._session_name: Optional[str] = None
+        self._base_docker_image: Optional[str] = None
         self._requirements: Optional[str] = None
         self._initial_scale: List[Dict[str, float]] = []
+        self._autosuspend_timeout = DEFAULT_AUTOSUSPEND_TIMEOUT
 
         # Whether to update the session when connecting to a fixed session.
         self._needs_update: bool = True
@@ -167,6 +196,18 @@ class SessionBuilder:
         # TODO(ekl) check for duplicate connections
         self._log.zero_time()
 
+        if ray.util.client.ray.is_connected():
+            raise RuntimeError(
+                "Already connected to a Ray session, please "
+                "run anyscale.connect in a new Python process."
+            )
+
+        # Allow the script to be run on the head node as well.
+        if "ANYSCALE_SESSION_ID" in os.environ:
+            self._log.info("Running in remote Anyscale session.")
+            ray.init(address="auto")
+            return
+
         # Autodetect or create a scratch project.
         if self._project_dir is None:
             self._project_dir = anyscale.project.find_project_root(os.getcwd())
@@ -180,7 +221,9 @@ class SessionBuilder:
 
         # TODO(ekl): generate a serverless compute configuraton here.
         cluster_yaml = yaml.safe_load(anyscale.project.CLUSTER_YAML_TEMPLATE)
-        self._populate_cluster_config(cluster_yaml)
+        self._populate_cluster_config(
+            cluster_yaml, self._anyscale_sdk.get_project(project_id).result.name
+        )
         with open(os.path.join(self._project_dir, "session-default.yaml"), "w+") as f:
             f.write(yaml.dump(cluster_yaml))
 
@@ -188,13 +231,15 @@ class SessionBuilder:
         # of the finger print.
         files_hash = self._fingerprint(proj_def.root)
 
-        if self._session_name is None or not self._get_session(
-            project_id, self._session_name
-        ):
-            # Unconditionally create the session if it doesn't exist.
-            needs_up = True
+        if self._session_name is not None:
+            sess = self._get_session(project_id, self._session_name)
+            if not sess or sess.state != "Running":
+                # Unconditionally create the session if it isn't up.
+                needs_up = True
+            else:
+                needs_up = self._needs_update
         else:
-            needs_up = self._needs_update
+            needs_up = False
 
         # Locate a updated session and update it if needed.
         if self._session_name is None or needs_up:
@@ -215,6 +260,31 @@ class SessionBuilder:
             self._tmp_conn.disconnect()
             self._tmp_conn = None
 
+        # Define ray in the notebook automatically for convenience.
+        try:
+            fr: Any = inspect.getouterframes(inspect.currentframe())[-1]
+            if fr.filename == "<stdin>" and "ray" not in fr.frame.f_globals:
+                self._log.debug("Auto importing Ray into the notebook.")
+                fr.frame.f_globals["ray"] = ray
+        except Exception as e:
+            self._log.error("Failed to auto define `ray` in notebook", e)
+
+    def cloud(self, cloud_name: str) -> "SessionBuilder":
+        """Set the name of the cloud to be used.
+
+        This sets the name of the cloud that your connect session will be
+        started in. If you do not specify it, it will use the last used cloud
+        in this project.
+
+        Args:
+            cloud_name (str): Name of the cloud to start the session in.
+
+        Examples:
+            >>> anyscale.cloud("aws_test_account").connect()
+        """
+        self._cloud_name = cloud_name
+        return self
+
     def project_dir(self, local_dir: str) -> "SessionBuilder":
         """Set the project directory.
 
@@ -263,7 +333,8 @@ class SessionBuilder:
         Examples:
             >>> anyscale.base_docker_image("anyscale/ray-ml:latest").connect()
         """
-        raise NotImplementedError()
+        self._base_docker_image = image_name
+        return self
 
     def require(self, requirements: Union[str, List[str]]) -> "SessionBuilder":
         """Set the Python requirements for the session.
@@ -332,26 +403,42 @@ class SessionBuilder:
         """
         raise NotImplementedError()
 
-    def autoshutdown(
-        self, enabled: bool = True, *, hours: int = 1, minutes: int = 0,
+    def autosuspend(
+        self,
+        enabled: bool = True,
+        *,
+        hours: Optional[int] = None,
+        minutes: Optional[int] = None,
     ) -> "SessionBuilder":
-        """Configure or disable session autoshutdown behavior.
+        """Configure or disable session autosuspend behavior.
 
-        The session will be autoshutdown after the specified time period. By
+        The session will be autosuspend after the specified time period. By
         default, sessions auto terminate after one hour of idle.
 
         Args:
-            enabled (bool): whether autoshutdown is enabled.
+            enabled (bool): whether autosuspend is enabled.
             hours (int): specify idle time in hours.
             minutes (int): specify idle time in minutes. This is added to the
                 idle time in hours.
 
         Examples:
-            >>> anyscale.autoshutdown(False).connect()
-            >>> anyscale.autoshutdown(hours=10).connect()
-            >>> anyscale.autoshutdown(hours=1, minutes=30).connect()
+            >>> anyscale.autosuspend(False).connect()
+            >>> anyscale.autosuspend(hours=10).connect()
+            >>> anyscale.autosuspend(hours=1, minutes=30).connect()
         """
-        raise NotImplementedError()
+        if enabled:
+            if hours is None and minutes is None:
+                timeout = DEFAULT_AUTOSUSPEND_TIMEOUT
+            else:
+                timeout = 0
+                if hours is not None:
+                    timeout += hours * 60
+                if minutes is not None:
+                    timeout += minutes
+        else:
+            timeout = -1
+        self._autosuspend_timeout = timeout
+        return self
 
     def nightly_build(self, git_commit: str) -> "SessionBuilder":
         """Use the specified nightly build commit for the session runtime.
@@ -420,9 +507,29 @@ class SessionBuilder:
     def _get_or_create_scratch_project(self) -> str:
         """Get or create a scratch project, including the directory."""
         project_dir = os.path.expanduser(self._scratch_dir)
-        self._ensure_project_setup_at_dir(project_dir)
+        project_name = os.path.basename(self._scratch_dir)
+        if not os.path.exists(project_dir) and self._project_exists(project_name):
+            self._clone_project(project_dir, project_name)
+        else:
+            self._ensure_project_setup_at_dir(project_dir)
         return project_dir
 
+    def _project_exists(self, project_name: str) -> bool:
+        """Return if a project of a given name exists."""
+        resp = self._anyscale_sdk.search_projects({"name": {"equals": project_name}})
+        return len(resp.results) > 0
+
+    def _clone_project(self, project_dir: str, project_name: str) -> None:
+        """Clone a project into the given dir by name."""
+        cur_dir = os.getcwd()
+        try:
+            parent_dir = os.path.dirname(project_dir)
+            os.makedirs(parent_dir, exist_ok=True)
+            os.chdir(parent_dir)
+            self._subprocess.check_call(["anyscale", "clone", project_name])
+        finally:
+            os.chdir(cur_dir)
+
     def _ensure_project_setup_at_dir(self, project_dir: str) -> None:
         """Get or create an Anyscale project rooted at the given dir."""
         os.makedirs(project_dir, exist_ok=True)
@@ -446,6 +553,90 @@ class SessionBuilder:
             with open(project_yaml, "w+") as f:
                 f.write(yaml.dump({"project_id": project_id}))
 
+    def _up_session(
+        self, project_id: str, session_name: str, cloud_name: str, cwd: Optional[str]
+    ) -> None:
+        # Non-blocking version of check_call, see
+        # https://github.com/python/cpython/blob/64abf373444944a240274a9b6d66d1cb01ecfcdd/Lib/subprocess.py#L363
+        command = [
+            "anyscale",
+            "up",
+            "--idle-timeout",
+            str(self._autosuspend_timeout),
+            "--config",
+            "session-default.yaml",
+            "--cloud-name",
+            cloud_name,
+            session_name,
+        ]
+        with tempfile.NamedTemporaryFile(delete=False) as output_log_file:
+            with self._subprocess.Popen(
+                command, cwd=cwd, stdout=output_log_file, stderr=subprocess.STDOUT,
+            ) as p:
+                # Get session URL:
+                session_found = False
+                retry_time = 1.0
+                while not session_found and retry_time < 10:
+                    results = self._list_sessions(project_id=project_id)
+                    for session in results:
+                        if session.name == session_name:
+                            self._log.info(
+                                "Updating session, see: {}/o/anyscale-internal/"
+                                "projects/{}/sessions/{}".format(
+                                    anyscale.conf.ANYSCALE_HOST, project_id, session.id
+                                )
+                            )
+                            session_found = True
+                    time.sleep(retry_time)
+                    retry_time = 2 * retry_time
+                try:
+                    retcode = p.wait()
+                except Exception as e:
+                    p.kill()
+                    raise e
+                # Check for errors:
+                if retcode:
+                    cmd = " ".join(command)
+                    msg = "Executing '{}' in {} failed.".format(cmd, self._project_dir)
+                    self._log.error(
+                        "--------------- Start update logs ---------------\n"
+                        "{}".format(open(output_log_file.name).read())
+                    )
+                    self._log.error("--------------- End update logs ---------------")
+                    raise RuntimeError(msg)
+
+    def _get_last_used_cloud(self, project_id: str) -> str:
+        """Return the name of the cloud last used in the project.
+
+        Args:
+            project_id (str): The project to get the last used cloud for.
+
+        Returns:
+            Name of the cloud last used in this project.
+        """
+        cloud_id = self._anyscale_sdk.get_project(project_id).result.last_used_cloud_id
+        if not cloud_id:
+            msg = (
+                "No last active cloud for this project yet. Use "
+                "anyscale.cloud('...').connect() to specify a cloud."
+            )
+            self._log.error(msg)
+            raise RuntimeError(msg)
+
+        # TODO(pcm): Replace this with an API call once the AnyscaleSDK supports it.
+        p = self._subprocess.Popen(
+            ["anyscale", "list", "clouds", "--json"], stdout=subprocess.PIPE
+        )
+        data = json.loads(p.communicate()[0])
+        cloud_name = {elem["id"]: elem["name"] for elem in data}[cloud_id]
+        self._log.info(
+            (
+                f"Using last active cloud '{cloud_name}'. "
+                "Call anyscale.cloud('...').connect() to overwrite."
+            )
+        )
+        return cast(str, cloud_name)
+
     def _get_or_create_updated_session(  # noqa: C901
         self, files_hash: bytes, project_id: str, session_name: Optional[str]
     ) -> str:
@@ -497,9 +688,9 @@ class SessionBuilder:
                     self._tmp_conn = self._acquire_session_lock(
                         session, raise_connection_error=True, connection_retries=0
                     )
-                except ConnectionError:
+                except Exception:
                     self._log.debug(
-                        "ConnectionError to session, will revisit", session.name
+                        "Error connecting to session, will revisit", session.name
                     )
                     session_name_conn_failed = session.name
                     continue
@@ -519,33 +710,28 @@ class SessionBuilder:
                 name = "session-{}".format(i)
                 if name not in used_names:
                     session_name = name
-                    self._log.info("Starting session", session_name)
+                    self._log.debug("Starting session", session_name)
                     break
 
         # Should not happen.
         if session_name is None:
             raise RuntimeError("Could not create new session to connect to.")
 
+        if self._cloud_name is None:
+            self._cloud_name = self._get_last_used_cloud(project_id)
+
         self._log.debug("Session hash", session_hash)
         self._log.debug("Local files hash", files_hash)
 
         if session_hash != files_hash:
-            self._log.info("Syncing latest project files to", session_name)
+            self._log.debug("Syncing latest project files to", session_name)
             # TODO(ekl): race condition here since "up" breaks the lock.
             if self._tmp_conn:
                 self._tmp_conn.disconnect()
                 self._tmp_conn = None
-            self._subprocess.check_call(
-                [
-                    "anyscale",
-                    "_up",
-                    "--config",
-                    "session-default.yaml",
-                    "--cloud-name",
-                    "anyscale_default_cloud",
-                    session_name,
-                ],
-                cwd=self._project_dir,
+            # Update session.
+            self._up_session(
+                project_id, session_name, self._cloud_name, self._project_dir
             )
             # Need to re-acquire the connection after the update.
             self._tmp_conn = self._acquire_session_lock(
@@ -573,7 +759,8 @@ class SessionBuilder:
         """Connect to and acquire a lock on the session.
 
         The session lock is released by calling disconnect() on the returned
-        Ray connection.
+        Ray connection. This function also checks for Python version
+        compatibility, it will not acquire the lock on version mismatch.
 
         Returns:
             Connection to the session, or None if a lock could not be acquired.
@@ -612,17 +799,6 @@ class SessionBuilder:
             session_url, secure=secure, metadata=metadata, connection_retries=3
         )
         self._log.debug("Server info", conn_info)
-        if "python_version" in conn_info:
-            local_py = "{}.{}".format(sys.version_info[0], sys.version_info[1])
-            remote_py = conn_info["python_version"].rsplit(".", 1)[0]
-
-            # TODO(ekl) avoid containers with version mismatch.
-            if local_py != remote_py:
-                self._log.info(
-                    "Warning: server Python version {} does not match "
-                    "local version {}, this is likely to cause serialization "
-                    "errors.".format(remote_py, local_py)
-                )
 
         def func() -> str:
             return "Connected!"
@@ -630,9 +806,9 @@ class SessionBuilder:
         f_remote = self._ray.remote(func)
         self._log.debug(self._ray.get(f_remote.remote()))
         self._log.info(
-            "View session at: {}/o/anyscale-internal/"
+            "Connected to {}, see: {}/o/anyscale-internal/"
             "projects/{}/sessions/{}".format(
-                anyscale.conf.ANYSCALE_HOST, project_id, session_found.id
+                session_name, anyscale.conf.ANYSCALE_HOST, project_id, session_found.id
             )
         )
 
@@ -685,9 +861,20 @@ class SessionBuilder:
                 fingerprint_hasher.update(f.read().encode("utf-8"))
         return fingerprint_hasher.hexdigest()
 
-    def _populate_cluster_config(self, cluster_yaml: Dict[str, Any]) -> None:
+    def _populate_cluster_config(
+        self, cluster_yaml: Dict[str, Any], project_name: str
+    ) -> None:
         """Populate cluster config with serverless compute options."""
 
+        base_docker_image_cpu = _get_base_image("cpu")
+        base_docker_image_gpu = _get_base_image("gpu")
+
+        # Overwrite base_docker_image if provided by user.
+        if self._base_docker_image:
+            base_docker_image_cpu = self._base_docker_image
+            base_docker_image_gpu = self._base_docker_image
+        self._log.debug("Base docker image {}".format(base_docker_image_cpu))
+
         # Setup serverless compute template.
         cluster_yaml["available_node_types"] = {
             "anyscale.cpu.medium": {
@@ -695,37 +882,38 @@ class SessionBuilder:
                 "max_workers": 20,
                 # TODO(ekl) why isn't resource autodetect working?
                 "resources": {"CPU": 16},
-                "docker": {"image": "anyscale/ray-ml:nightly-cpu"},
+                "docker": {"image": base_docker_image_cpu},
             },
             "anyscale.cpu.large": {
                 "node_config": {"InstanceType": "m5.16xlarge"},
                 "max_workers": 10,
                 "resources": {"CPU": 64},
-                "docker": {"image": "anyscale/ray-ml:nightly-cpu"},
+                "docker": {"image": base_docker_image_cpu},
             },
             "anyscale.gpu.medium": {
                 "node_config": {"InstanceType": "g3.4xlarge"},
                 "max_workers": 20,
                 "resources": {"CPU": 16, "GPU": 1},
-                "docker": {"image": "anyscale/ray-ml:nightly-gpu"},
+                "docker": {"image": base_docker_image_gpu},
             },
             "anyscale.gpu.large": {
                 "node_config": {"InstanceType": "g3.16xlarge"},
                 "max_workers": 10,
                 "resources": {"CPU": 64, "GPU": 4},
-                "docker": {"image": "anyscale/ray-ml:nightly-gpu"},
+                "docker": {"image": base_docker_image_gpu},
             },
         }
         cluster_yaml["max_workers"] = 50
         cluster_yaml["head_node_type"] = "anyscale.cpu.medium"
         cluster_yaml["worker_default_node_type"] = "anyscale.cpu.medium"
         # TODO(ekl) this shouldn't be needed.
-        cluster_yaml["docker"]["image"] = "anyscale/ray-ml:nightly-gpu"
+        cluster_yaml["docker"]["image"] = base_docker_image_gpu
 
-        # Install Ray runtime.
-        cluster_yaml["setup_commands"] = [
-            "pip install -U {}".format(self._get_wheel_url(self._ray_release))
-        ]
+        # Install Ray runtime if specified.
+        if self._ray_release:
+            cluster_yaml["setup_commands"] = [
+                "pip install -U {}".format(self._get_wheel_url(self._ray_release))
+            ]
 
         # Install requirements:
         if self._requirements:
@@ -733,11 +921,11 @@ class SessionBuilder:
                 f'echo "{self._requirements}" | pip install -r /dev/stdin'
             )
 
-        # TODO(ekl) ray client server should be started by default
+        # TODO(ekl) we should make the `cd` here standard for all sessions.
         cluster_yaml["head_start_ray_commands"] = [
             "ray stop",
             "ulimit -n 65536; cd ~/"
-            + os.path.basename(str(self._project_dir))
+            + project_name
             + "; AUTOSCALER_UPDATE_INTERVAL_S=5 ray start --head --port=6379 "
             "--object-manager-port=8076 "
             "--ray-client-server-port=10001 "
@@ -753,6 +941,9 @@ class SessionBuilder:
             )
         )
 
+    def _get_project_name(self, project_id: str) -> str:
+        return self._anyscale_sdk.get_project(project_id).result.name  # type: ignore
+
     def _list_sessions(self, project_id: str) -> List[Session]:
         sessions = []
         has_more = True
