@@ -1,4 +1,5 @@
 import copy
+from datetime import datetime
 import json
 import os
 import subprocess
@@ -18,7 +19,17 @@ from anyscale.autosync_heartbeat import (
     perform_autopush_synchronization,
     perform_sync,
 )
-from anyscale.client.openapi_client import CreateSessionFromSnapshotOptions, Session, SessionFinishUpOptions, SessionState, SessionUpOptions, SetupInitializeSessionOptions, StopSessionOptions  # type: ignore
+from anyscale.client.openapi_client import (  # type: ignore
+    ComputeTemplateConfig,
+    CreateComputeTemplate,
+    CreateSessionFromSnapshotOptions,
+    Session,
+    SessionFinishUpOptions,
+    SessionState,
+    SessionUpOptions,
+    SetupInitializeSessionOptions,
+    StopSessionOptions,
+)
 from anyscale.client.openapi_client.api.default_api import DefaultApi  # type: ignore
 from anyscale.cloud import get_cloud_id_and_name
 from anyscale.cluster_config import (
@@ -336,7 +347,7 @@ class SessionController:
         session_name: Optional[str],
         config: Optional[str],
         build_id: Optional[str],
-        compute_template_id: Optional[str],
+        compute_config: Optional[str],
         min_workers: Optional[int],
         max_workers: Optional[int],
         no_restart: bool,
@@ -346,6 +357,7 @@ class SessionController:
         cloud_name: Optional[str],
         idle_timeout: Optional[int],
         verbose: bool = False,
+        no_rapid_start: bool = False,
     ) -> None:
         # TODO(ilr) Split this into ~3 functions, look at jbai's comments on #1469
         project_definition = load_project_or_throw()
@@ -373,7 +385,7 @@ class SessionController:
 
         # Old cluster config launch path.
         cluster_config = None
-        if build_id is None and compute_template_id is None:
+        if build_id is None and compute_config is None:
             if session_exists and not config:
                 # Get cluster config stored in DB from existing session
                 session = session_list[0]
@@ -400,6 +412,8 @@ class SessionController:
             cluster_config = _configure_min_max_workers(
                 cluster_config, min_workers, max_workers
             )
+            if no_rapid_start:
+                cluster_config["provider"]["bypass_rapid_start"] = True
 
         # Build and compute template launch path.
         else:
@@ -423,47 +437,90 @@ class SessionController:
         if idle_timeout is None:
             idle_timeout = 120
 
-        with format_api_exception(ApiException):
-            up_response = self.api_client.start_empty_session_api_v2_sessions_start_empty_session_post(
-                session_up_options=SessionUpOptions(
+        # Register the compute config with Anyscale and get its ID.
+        compute_template_id = None
+        if compute_config:
+            with open(compute_config, "r") as f:
+                config_dict = json.load(f)
+            config_object = ComputeTemplateConfig(**config_dict)
+            created_template = self.api_client.create_compute_template_api_v2_compute_templates_post(
+                create_compute_template=CreateComputeTemplate(
+                    name="autogenerated-config-{}".format(datetime.now().isoformat()),
                     project_id=project_id,
-                    name=session_name,
-                    cluster_config={"config": json.dumps(cluster_config)}
-                    if cluster_config
-                    else None,
-                    build_id=build_id,
-                    compute_template_id=compute_template_id,
-                    cloud_id=cloud_id,
-                    idle_timeout=idle_timeout,
-                    restart_only=restart_only,
-                    no_restart=no_restart,
+                    config=config_object,
                 )
             ).result
-        session_id = up_response.session_id
-        second_update_required = up_response.second_update_required
-        wait_for_head_node_start(project_id, session_name, self.api_client)
+            compute_template_id = created_template.id
 
-        # Push will sync project directory to working directory.
-        self.push(
-            session_name, source=None, target=None, config=None, all_nodes=False,
-        )
-
-        with format_api_exception(ApiException):
-            self.api_client.setup_and_initialize_session_api_v2_sessions_session_id_setup_and_initialize_session_post(
-                session_id=session_id,
-                setup_initialize_session_options=SetupInitializeSessionOptions(
-                    project_id=project_id,
-                    cluster_config={"config": json.dumps(cluster_config)},
-                    restart_only=restart_only,
-                    no_restart=no_restart,
-                    no_update=not second_update_required,
-                ),
+        try:
+            with format_api_exception(ApiException):
+                up_response = self.api_client.start_empty_session_api_v2_sessions_start_empty_session_post(
+                    session_up_options=SessionUpOptions(
+                        project_id=project_id,
+                        name=session_name,
+                        cluster_config={"config": json.dumps(cluster_config)}
+                        if cluster_config
+                        else None,
+                        build_id=build_id,
+                        compute_template_id=compute_template_id,
+                        cloud_id=cloud_id,
+                        idle_timeout=idle_timeout,
+                        restart_only=restart_only,
+                        no_restart=no_restart,
+                    )
+                ).result
+            session_id = up_response.session_id
+            second_update_required = up_response.second_update_required
+            wait_for_head_node_start(
+                project_id, session_name, session_id, self.api_client
             )
 
-        wait_for_session_start(project_id, session_name, self.api_client)
+            try:
+                # Push will sync project directory to working directory.
+                self.push(
+                    session_name,
+                    source=None,
+                    target=None,
+                    config=None,
+                    all_nodes=False,
+                )
+            except Exception as e:
+                self.api_client.cancel_startup_api_v2_sessions_session_id_cancel_startup_post(
+                    session_id=session_id
+                )
+                raise click.ClickException(
+                    "{}\nSession startup failed while transferring files".format(e)
+                )
 
-        url = get_endpoint(f"/projects/{project_id}")
-        print(f"Session {session_name} started. View at {url}")
+            with format_api_exception(ApiException):
+                self.api_client.setup_and_initialize_session_api_v2_sessions_session_id_setup_and_initialize_session_post(
+                    session_id=session_id,
+                    setup_initialize_session_options=SetupInitializeSessionOptions(
+                        project_id=project_id,
+                        cluster_config={"config": json.dumps(cluster_config)},
+                        restart_only=restart_only,
+                        no_restart=no_restart,
+                        no_update=not second_update_required,
+                    ),
+                )
+
+            wait_for_session_start(project_id, session_name, self.api_client)
+            url = get_endpoint(f"/projects/{project_id}/sessions/{session_id}")
+            print(f"Session {session_name} started. View at {url}")
+        except KeyboardInterrupt:
+            self.api_client.stop_session_api_v2_sessions_session_id_stop_post(
+                session_id,
+                StopSessionOptions(
+                    terminate=True,
+                    workers_only=False,
+                    keep_min_workers=False,
+                    delete=False,
+                    take_snapshot=False,
+                ),
+            )
+            raise click.ClickException(
+                f"Terminating session {session_name} due to keyboard interrupt."
+            )
 
     def fork_session(
         self,
@@ -717,7 +774,7 @@ class SessionController:
                 idle_timeout=None,
                 yes=True,
             )
-        url = get_endpoint(f"/projects/{project_id}")
+        url = get_endpoint(f"/projects/{project_id}/sessions/{session.id}")
         print(f"Pushed to session {session_name}. View at {url}")
 
     # Helpers
