@@ -6,12 +6,10 @@ from typing import Any
 
 import click
 from packaging import version
-import ray.ray_constants
 
 from anyscale.api import instantiate_api_client
 from anyscale.client.openapi_client.rest import ApiException  # type: ignore
-from anyscale.cloudgateway import CloudGatewayRunner
-from anyscale.commands.anyscale_api.api_commands import anyscale_api  # type: ignore
+from anyscale.commands.anyscale_api.api_commands import anyscale_api
 from anyscale.commands.cloud_commands import cloud_cli
 from anyscale.commands.exec_commands import anyscale_exec
 from anyscale.commands.list_commands import list_cli
@@ -25,6 +23,7 @@ from anyscale.commands.session_commands import (
     anyscale_ssh,
     anyscale_stop,
     anyscale_up,
+    anyscale_up_internal,
 )
 from anyscale.commands.session_commands_hidden import (
     anyscale_start,
@@ -35,7 +34,6 @@ import anyscale.conf
 from anyscale.util import send_json_request
 
 
-logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
 logger = logging.getLogger(__file__)
 logging.getLogger("botocore").setLevel(logging.CRITICAL)
 
@@ -77,6 +75,10 @@ class AliasedGroup(click.Group):
 )
 @click.pass_context
 def cli(ctx: Any, version_flag: bool, show_json: bool) -> None:
+
+    # Use anyscale vendored Ray for the CLI:
+    sys.path.insert(0, anyscale.ANYSCALE_RAY_DIR)
+
     try:
         api_client = instantiate_api_client(no_cli_token=True)
         resp = api_client.get_anyscale_version_api_v2_userinfo_anyscale_version_get()
@@ -137,6 +139,8 @@ def anyscale_help(ctx: Any) -> None:
 )
 @click.option("--gateway-id", type=str, required=True)
 def anyscale_cloudgateway(gateway_id: str) -> None:
+    from anyscale.cloudgateway import CloudGatewayRunner
+
     # Make sure only registered users can start the gateway.
     logger.info("Verifying user ...")
     try:
@@ -170,6 +174,7 @@ cli.add_command(list_cli)
 cli.add_command(anyscale_init)
 cli.add_command(anyscale_start)
 cli.add_command(anyscale_up)
+cli.add_command(anyscale_up_internal)
 cli.add_command(anyscale_stop)
 cli.add_command(anyscale_cloudgateway)
 cli.add_command(anyscale_autosync)
