@@ -5,7 +5,6 @@ import shutil
 from typing import Any, List, Optional
 
 import click
-import ray.ray_constants
 import requests
 
 from anyscale.client.openapi_client.rest import ApiException  # type: ignore
@@ -17,7 +16,6 @@ from anyscale.util import (
 )
 
 
-logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
 logger = logging.getLogger(__file__)
 
 
