@@ -15,6 +15,7 @@ from anyscale.client.openapi_client import (  # type: ignore
     SessionStartingUpData,
     SessionStateData,
 )
+from backend.server.base_api.models.session_commands import SessionCommandTypes
 
 
 @pytest.fixture()
@@ -60,6 +61,8 @@ def project_test_data() -> Project:
         creator_id="creator_id",
         is_owner=True,
         directory_name="/directory/name",
+        active_sessions=0,
+        last_activity_at=datetime.now(tz=timezone.utc),
     )
 
 
@@ -117,6 +120,7 @@ def session_command_test_data() -> SessionCommand:
         params="params",
         shell="shell",
         shell_command="shell_command",
+        type=SessionCommandTypes.COMMAND_LINE_RUNNER,
     )
 
 
