@@ -7,7 +7,10 @@ from anyscale.client.openapi_client import Project  # type: ignore
 from anyscale.client.openapi_client.models.project_list_response import (  # type: ignore
     ProjectListResponse,
 )
-from anyscale.controllers.project_controller import ProjectController
+from anyscale.controllers.project_controller import (
+    COMPUTE_CONFIG_FILENAME,
+    ProjectController,
+)
 from anyscale.project import ProjectDefinition
 
 
@@ -28,6 +31,7 @@ def mock_api_client(project_test_data: Project) -> Mock:
 
 def test_clone_project(project_test_data: Project, mock_api_client: Mock) -> None:
     project_controller = ProjectController(api_client=mock_api_client)
+    project_controller._write_sample_compute_config = Mock()  # type: ignore
 
     os_makedirs_mock = Mock(return_value=None)
     with patch.multiple("os", makedirs=os_makedirs_mock), patch(
@@ -46,12 +50,17 @@ def test_clone_project(project_test_data: Project, mock_api_client: Mock) -> Non
         project_test_data.id, "", project_test_data.name
     )
 
+    project_controller._write_sample_compute_config.assert_called_once_with(
+        os.path.join(project_test_data.name, COMPUTE_CONFIG_FILENAME)
+    )
+
 
 @pytest.mark.parametrize("config", [None, "tmp.yaml"])
 def test_init_project(
     project_test_data: Project, mock_api_client: Mock, config: str
 ) -> None:
     project_controller = ProjectController(api_client=mock_api_client)
+    project_controller._write_sample_compute_config = Mock()  # type: ignore
     project_definition = ProjectDefinition(os.getcwd())
     project_definition.config["name"] = project_test_data.name
     mock_create_new_proj_def = Mock(
@@ -80,3 +89,7 @@ def test_init_project(
         mock_validate_cluster_configuration.assert_called_once_with(
             config, api_instance=mock_api_client
         )
+
+    project_controller._write_sample_compute_config.assert_called_once_with(
+        COMPUTE_CONFIG_FILENAME
+    )
