@@ -84,8 +84,8 @@ def mock_api_client_multiple_sessions(base_mock_api_client: Mock) -> Mock:
                 tensorboard_available=False,
                 project_id="project_id",
                 state="Running",
-                last_activity_at=datetime.now(tz=timezone.utc),
                 head_node_ip="127.0.0.1",
+                idle_timeout=120,
             ),
             Session(
                 id="ses_2",
@@ -95,8 +95,8 @@ def mock_api_client_multiple_sessions(base_mock_api_client: Mock) -> Mock:
                 tensorboard_available=False,
                 project_id="project_id",
                 state="Running",
-                last_activity_at=datetime.now(tz=timezone.utc),
                 head_node_ip="127.0.0.1",
+                idle_timeout=120,
             ),
         ]
     )
@@ -227,6 +227,8 @@ def test_anyscale_up_success(mock_up_tuple: Tuple[Mock, Project, Cloud, Mock]) -
         session_controller.up(
             session_name=None,
             config=temp_file.name,
+            build_id=None,
+            compute_template_id=None,
             min_workers=None,
             max_workers=None,
             no_restart=False,
@@ -234,6 +236,7 @@ def test_anyscale_up_success(mock_up_tuple: Tuple[Mock, Project, Cloud, Mock]) -
             disable_sync=False,
             cloud_id=None,
             cloud_name=None,
+            idle_timeout=-1,
             yes=True,
         )
         mock_api_client.session_up_api_v2_sessions_up_post.assert_called_with(
@@ -244,12 +247,144 @@ def test_anyscale_up_success(mock_up_tuple: Tuple[Mock, Project, Cloud, Mock]) -
                     "config": json.dumps(yaml.safe_load(CLUSTER_YAML_TEMPLATE))
                 },
                 cloud_id=cloud_test_data.id,
+                idle_timeout=-1,
             )
         )
 
     mock_api_client.session_up_api_v2_sessions_up_post.assert_called_once()
 
 
+def test_anyscale_up_templates_success(
+    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock]
+) -> None:
+    (
+        mock_api_client,
+        project_test_data,
+        cloud_test_data,
+        mock_wait_for_start,
+    ) = mock_up_tuple
+
+    session_controller = SessionController(api_client=mock_api_client)
+    mock_api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get.return_value = ProjectdefaultsessionnameResponse(
+        result=ProjectDefaultSessionName(name="session-11")
+    )
+    mock_api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
+        results=[]  # No session with this name yet, brand new session.
+    )
+
+    mock_api_client.session_up_api_v2_sessions_up_post = Mock(
+        return_value=SessionupresponseResponse(
+            result=SessionUpResponse(
+                session_id="session-10",
+                cluster_name="new_cluster",
+                private_key="key",
+                cluster_config=yaml.safe_load(CLUSTER_YAML_TEMPLATE),
+                is_same_cluster_config=False,
+            )
+        )
+    )
+
+    session_controller.up(
+        session_name=None,
+        config=None,
+        build_id="build-id",
+        compute_template_id="ct-id",
+        min_workers=None,
+        max_workers=None,
+        idle_timeout=-1,
+        no_restart=False,
+        restart_only=False,
+        disable_sync=False,
+        cloud_id=None,
+        cloud_name=None,
+        yes=True,
+    )
+    # Ensure that the API call is made with the
+    # build ID and compute template ID,
+    # and not the cluster config.
+    mock_api_client.session_up_api_v2_sessions_up_post.assert_called_with(
+        session_up_options=SessionUpOptions(
+            project_id=project_test_data.id,
+            name="session-11",
+            cluster_config=None,
+            idle_timeout=-1,
+            build_id="build-id",
+            compute_template_id="ct-id",
+            cloud_id=cloud_test_data.id,
+        )
+    )
+    mock_api_client.session_up_api_v2_sessions_up_post.assert_called_once()
+
+
+def test_anyscale_up_templates_badparams(
+    mock_up_tuple: Tuple[Mock, Project, Cloud, Mock]
+) -> None:
+    (
+        mock_api_client,
+        project_test_data,
+        cloud_test_data,
+        mock_wait_for_start,
+    ) = mock_up_tuple
+
+    session_controller = SessionController(api_client=mock_api_client)
+    mock_api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get.return_value = ProjectdefaultsessionnameResponse(
+        result=ProjectDefaultSessionName(name="session-11")
+    )
+    mock_api_client.list_sessions_api_v2_sessions_get.return_value = SessionListResponse(
+        results=[]  # No session with this name yet, brand new session.
+    )
+
+    mock_api_client.session_up_api_v2_sessions_up_post = Mock(
+        return_value=SessionupresponseResponse(
+            result=SessionUpResponse(
+                session_id="session-10",
+                cluster_name="new_cluster",
+                private_key="key",
+                cluster_config=yaml.safe_load(CLUSTER_YAML_TEMPLATE),
+                is_same_cluster_config=False,
+            )
+        )
+    )
+
+    # Specifying only the build ID or only the compute template ID
+    # should raise a ValueError.
+    # (If neither are specified, it counts as a cluster config
+    # startup path, and is handled in other existing tests.)
+    with pytest.raises(ValueError):
+        session_controller.up(
+            session_name=None,
+            config=None,
+            build_id="build-id",
+            compute_template_id=None,
+            min_workers=None,
+            max_workers=None,
+            idle_timeout=-1,
+            no_restart=False,
+            restart_only=False,
+            disable_sync=False,
+            cloud_id=None,
+            cloud_name=None,
+            yes=True,
+        )
+
+    with pytest.raises(ValueError):
+        session_controller.up(
+            session_name=None,
+            config=None,
+            build_id=None,
+            compute_template_id="ct-id",
+            min_workers=None,
+            max_workers=None,
+            idle_timeout=-1,
+            no_restart=False,
+            restart_only=False,
+            disable_sync=False,
+            cloud_id=None,
+            cloud_name=None,
+            yes=True,
+        )
+
+
 def test_anyscale_up_failure(mock_up_tuple: Tuple[Mock, Project, Cloud, Mock]) -> None:
     (
         mock_api_client,
@@ -291,6 +426,8 @@ def test_anyscale_up_failure(mock_up_tuple: Tuple[Mock, Project, Cloud, Mock]) -
             session_controller.up(
                 session_name=None,
                 config=temp_file.name,
+                build_id=None,
+                compute_template_id=None,
                 min_workers=None,
                 max_workers=None,
                 no_restart=False,
@@ -298,6 +435,7 @@ def test_anyscale_up_failure(mock_up_tuple: Tuple[Mock, Project, Cloud, Mock]) -
                 disable_sync=False,
                 cloud_id=None,
                 cloud_name=None,
+                idle_timeout=-1,
                 yes=True,
             )
         mock_api_client.session_finish_up_api_v2_sessions_session_id_finish_up_post.assert_called_once_with(
@@ -344,6 +482,8 @@ def test_anyscale_up_no_config_provided(
     session_controller.up(
         session_name=None,
         config=None,
+        build_id=None,
+        compute_template_id=None,
         min_workers=None,
         max_workers=None,
         no_restart=False,
@@ -351,6 +491,7 @@ def test_anyscale_up_no_config_provided(
         disable_sync=False,
         cloud_id=None,
         cloud_name=None,
+        idle_timeout=-1,
         yes=True,
     )
     mock_api_client.session_up_api_v2_sessions_up_post.assert_called_with(
@@ -359,6 +500,7 @@ def test_anyscale_up_no_config_provided(
             name="session-11",
             cluster_config=None,
             cloud_id=cloud_test_data.id,
+            idle_timeout=-1,
         )
     )
 
@@ -402,6 +544,8 @@ def test_anyscale_up_config_provided(
         session_controller.up(
             session_name=None,
             config=temp_file.name,
+            build_id=None,
+            compute_template_id=None,
             min_workers=None,
             max_workers=None,
             no_restart=False,
@@ -409,6 +553,7 @@ def test_anyscale_up_config_provided(
             disable_sync=False,
             cloud_id=None,
             cloud_name=None,
+            idle_timeout=-1,
             yes=True,
         )
         mock_api_client.session_up_api_v2_sessions_up_post.assert_called_with(
@@ -419,6 +564,7 @@ def test_anyscale_up_config_provided(
                     "config": json.dumps(yaml.safe_load(CLUSTER_YAML_TEMPLATE))
                 },
                 cloud_id=cloud_test_data.id,
+                idle_timeout=-1,
             )
         )
 
@@ -691,12 +837,7 @@ def test_anyscale_push_session(
         # Don't check the config_file argument because it's a random path of
         # a temporary file
         mock_rsync.assert_called_once_with(
-            ANY,
-            source=source,
-            target=target,
-            override_cluster_name=None,
-            down=False,
-            all_nodes=all_nodes,
+            ANY, source=source, target=target, down=False,
         )
 
         mock_validate_cluster_configuration.assert_called_once_with(
@@ -706,9 +847,12 @@ def test_anyscale_push_session(
         mock_up.assert_called_once_with(
             session_name=session_test_data.name,
             config=None,
+            build_id=None,
+            compute_template_id=None,
             min_workers=None,
             max_workers=None,
             no_restart=False,
+            idle_timeout=None,
             restart_only=False,
             disable_sync=True,
             cloud_id=session_test_data.cloud_id,
